# Copyright (C) 2009, 2013 Red Hat, Inc.
# Copyright (C) 2009 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import collections
import os
import shutil
import datetime
import subprocess
import gnupg
import libvirt

from crontab import CronTab

from gi.repository import Gtk
from gi.repository import Pango

from virtinst import Backuper
from virtinst import log
from virtinst import xmlutil

from .lib import uiutil
from .baseclass import vmmGObjectUI
from .asyncjob import vmmAsyncJob


def _get_exporable_msg(diskinfo):
    """Is the passed path even backup-able"""
    if diskinfo.get_exporable_msg():
        return diskinfo.get_exporable_msg()
    if diskinfo.disk.is_empty():
        return _("No storage to backup.")


class _StorageInfo:
    """
    Our wrapper class that is close to backuper _DiskInfo content.
    We track all user choices explicitly in this class, and then
    serialize them into the actual backuper diskinfo once the
    user clicks 'backup' to complete the process.
    """
    def __init__(self, vm, backuper, diskinfo):
        self._diskinfo = diskinfo
        self._orig_disk = diskinfo.disk
        self._orig_vm_name = vm.get_name()
        self._new_vm_name = backuper._new_name

        self.share_msg = diskinfo.get_share_msg()
        self.exporable_msg = _get_exporable_msg(diskinfo)

        self._manual_path = None
        self._generated_path = (diskinfo.new_disk and
                diskinfo.new_disk.get_source_path())
        self._is_clone_requested = diskinfo.is_clone_requested()

        if self.share_msg or not self.is_cloneable():
            self._is_clone_requested = False

    def is_cloneable(self):
        return not bool(self.exporable_msg)
    def is_clone_requested(self):
        return self._is_clone_requested
    def is_share_requested(self):
        return not self._is_clone_requested
    def warn_about_sharing(self):
        return not self.share_msg

    def get_target(self):
        return self._orig_disk.target
    def get_orig_disk_path(self):
        return self._orig_disk.get_source_path()
    def get_new_disk_path(self):
        if self._manual_path:
            return self._manual_path
        return self._generated_path
    def get_share_msg(self):
        return self._diskinfo.get_share_msg()
    def get_driver_type(self):
        return self._orig_disk.driver_type

    def set_clone_requested(self, val):
        self._is_clone_requested = bool(val)
    def set_manual_path(self, val):
        self._manual_path = val
    def set_new_vm_name(self, val):
        if not val or val == self._new_vm_name:
            return
        self._new_vm_name = val

    def _reset_generated_path(self):
        self._generated_path = Backuper.generate_backup_disk_path(
            self._orig_disk.conn,
            self._orig_vm_name,
            self._new_vm_name,
            self._orig_disk.get_source_path())

    def set_values_on_diskinfo(self, diskinfo):
        if not self._is_clone_requested:
            diskinfo.set_share_requested()
            diskinfo.new_disk = None
            return

        diskinfo.set_clone_requested()


    ###################
    # UI info helpers #
    ###################

    def get_tooltip(self):
        lines = []
        lines.append(_("Disk target: %s") % self.get_target())
        lines.append(_("Original path: %s") % self.get_orig_disk_path())
        if self.get_new_disk_path():
            lines.append(_("New path: %s") % self.get_new_disk_path())
        lines.append("\n")

        if self.share_msg:
            lines.append(_("Storage is safe to share: %(reason)s") % {
                "reason": self.share_msg})
        else:
            lines.append(
                _("Sharing this storage is potentially dangerous."))

        if self.exporable_msg:
            lines.append(_("Storage is not backupable: %(reason)s") % {
                "reason": self.exporable_msg})
        return "\n".join(lines)

    def get_markup(self, vm):
        lines = []

        line = ""
        path = self.get_orig_disk_path()
        if path:
            line += path
        else:
            line += _("No storage.")
        lines.append(line)

        line = ""
        if self.is_share_requested():
            line += _("Share disk with %s") % vm.get_name()
        if self.is_clone_requested():
            line += _("Backup this disk")
            sizelabel = self.get_size_label()
            if sizelabel:
                line += " (%s)" % sizelabel
        if line:
            lines.append(line)

        label = "\n".join(lines)
        markup = "<small>%s</small>" % xmlutil.xml_escape(label)
        return markup

    def get_icon_name(self):
        if self._orig_disk.is_floppy():
            return "media-floppy"
        if self._orig_disk.is_cdrom():
            return "media-optical"
        return "drive-harddisk"

    def get_size_label(self):
        size = self._orig_disk.get_size()
        if not size:
            return ""  # pragma: no cover
        return "%.1f GiB" % float(size)


class vmmBackupDialog(vmmGObjectUI):
    _instance = None

    @classmethod
    def reset_instance(cls):
        cls._instance = None

    @classmethod
    def show_instance(cls, parentobj, vm):
        try:
            if cls._instance:
                cls._instance.close()
            cls.reset_instance()
            cls._instance = vmmBackupDialog(vm)
            cls._instance.show(parentobj.topwin, vm)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(
                    _("Error launching backup dialog: %s") % str(e))

    def __init__(self, vm):
        vmmGObjectUI.__init__(self, "backup.ui", "vmm-backup")
        self.vm = vm
        self._storage_list = None
        self._encryption_key = None

        self._schedule_dialog = self.widget("vmm-backup-schedule")
        self._schedule_dialog.set_transient_for(self.topwin)
        self._encryption_dialog = self.widget("vmm-backup-encryption")
        self._encryption_dialog.set_transient_for(self.topwin)

        self.builder.connect_signals({
            "on_backup_delete_event": self._close_cb,
            "on_backup_delete_clicked": self._backup_delete_cb,
            "on_backup_cancel_clicked": self._close_cb,
            "on_backup_ok_clicked": self._finish_clicked_cb,
            "on_backup_schedule_clicked": self._schedule_backup_cb,
            "on_browse_local_clicked": self._browse_local_clicked_cb,
            "on_backup_folder_changed": self._backup_folder_changed_cb,
            "on_backup_encrypt_clicked": self._backup_encrypt_cb,

            # Schedule subdialog signals
            "on_vmm_backup_schedule_delete_event": self._schedule_dialog_close_cb,
            "on_backup_schedule_delete_clicked": self._schedule_delete_cb,
            "on_backup_schedule_cancel_clicked": self._schedule_dialog_close_cb,
            "on_backup_schedule_ok_clicked": self._schedule_dialog_finish_cb,

            # Encryption subdialog signals
            "on_vmm_backup_encryption_delete_event": self._encryption_dialog_close_cb,
            "on_backup_encryption_delete_clicked": self._encryption_delete_cb,
            "on_backup_encryption_cancel_clicked": self._encryption_dialog_close_cb,
            "on_backup_encryption_ok_clicked": self._encryption_dialog_finish_cb,
            "on_backup_encryption_key_toggled": self._encryption_key_toggled_cb,

            "on_threshold_toggled": self._threshold_toggled_cb,
            "on_compression_toggled": self._compression_toggled_cb,
        })
        self._init_ui()
        self._init_backup_data()


    #######################
    # Standard UI methods #
    #######################

    @property
    def conn(self):
        return self.vm and self.vm.conn or None

    def show(self, parent, vm):
        log.debug("Showing backup wizard")
        self._set_vm(vm)
        self._reset_state()
        self.topwin.set_transient_for(parent)
        self.topwin.resize(1, 1)
        self.topwin.present()

    def _schedule_dialog_close(self):
        self._schedule_dialog.hide()
        return 1

    def _encryption_dialog_close(self):
        self._encryption_dialog.hide()
        return 1

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing backup wizard")
        self._schedule_dialog_close()
        self.topwin.hide()

        self._set_vm(None)
        self._storage_list = None
        return 1

    def _vm_removed_cb(self, _conn, vm):
        if self.vm == vm:
            self.close()

    def _set_vm(self, newvm):
        oldvm = self.vm
        if oldvm:
            oldvm.conn.disconnect_by_obj(self)
        if newvm:
            newvm.conn.connect("vm-removed", self._vm_removed_cb)
        self.vm = newvm

    def _cleanup(self):
        self._schedule_dialog.destroy()
        self._schedule_dialog = None
        self._encryption_dialog.destroy()
        self._encryption_dialog = None
        self._encryption_key = None


    ###########
    # UI init #
    ###########

    def _init_ui(self):
        storage_list = self.widget("storage-list")

        # [disk target, tooltip]
        model = Gtk.ListStore(str, str)
        storage_list.set_model(model)
        storage_list.set_tooltip_column(1)

        backupCol = Gtk.TreeViewColumn(_("Backup"))
        pathCol = Gtk.TreeViewColumn(_("Storage"))
        pathCol.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        pathCol.set_expand(True)

        storage_list.append_column(backupCol)
        storage_list.append_column(pathCol)

        def separator_cb(_model, _iter):
            return _model[_iter][0] == "separator"
        storage_list.set_row_separator_func(separator_cb)

        chkbox = Gtk.CellRendererToggle()
        chkbox.connect('toggled', self._storage_backup_toggled_cb)
        chkimg = Gtk.CellRendererPixbuf()
        chkimg.set_property('stock-size', Gtk.IconSize.MENU)
        backupCol.pack_start(chkimg, False)
        backupCol.pack_start(chkbox, False)

        def chk_cell_data_cb(column, cell, model, _iter, data):
            target = model[_iter][0]
            if target == "separator":
                return
            sinfo = self._storage_list[target]
            visible = sinfo.is_cloneable()
            active = sinfo.is_clone_requested()
            _chkimg = column.get_cells()[0]
            _chkbox = column.get_cells()[1]
            _chkbox.set_property('active', active)
            _chkbox.set_property('visible', visible)
            _chkimg.set_property('visible', not visible)
            icon = "dialog-information"
            if sinfo.warn_about_sharing():
                icon = "dialog-warning"
            _chkimg.set_property('icon-name', icon)
            tooltip = sinfo.get_tooltip()
            if tooltip != model[_iter][1]:
                model[_iter][1] = tooltip
            if sinfo.get_driver_type() == "raw":
                self.widget("raw-disk-notification-box").set_visible(True)

        backupCol.set_cell_data_func(chkbox, chk_cell_data_cb)
        backupCol.set_cell_data_func(chkimg, chk_cell_data_cb)

        pathtxt = Gtk.CellRendererText()
        pathCol.set_sort_column_id(0)
        pathtxt.set_property("width-chars", 30)
        pathtxt.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)
        pathimg = Gtk.CellRendererPixbuf()
        pathimg.set_property('stock-size', Gtk.IconSize.MENU)
        pathimg.set_padding(3, 0)
        pathCol.pack_start(pathimg, False)
        pathCol.pack_start(pathtxt, True)

        def path_cb(column, cell, model, _iter, data):
            target = model[_iter][0]
            if target == "separator":
                return
            sinfo = self._storage_list[target]
            _pathimg = column.get_cells()[0]
            _pathtxt = column.get_cells()[1]
            markup = sinfo.get_markup(self.vm)
            _pathtxt.set_property('markup', markup)
            _pathimg.set_property('icon-name', sinfo.get_icon_name())

        pathCol.set_cell_data_func(pathtxt, path_cb)
        pathCol.set_cell_data_func(pathimg, path_cb)

        w = self.widget
        self._populate_repeat_combo(w("schedule-repeat"))
        self._populate_backup_type_combo(w("backup-type"))
        self._populate_encryption_key_combo(w("encryption-key"))
        self._populate_hash_combo(w("backup-hash"))

        if self.vm.get_xmlobj(inactive=True).get_backup_folder():
            w("backup-delete").set_visible(True)
            w("backup-cancel").set_visible(False)

        tooltip = self._cron_tooltip()
        w("schedule-button").set_tooltip_text(tooltip)
        w("schedule-button").set_sensitive(tooltip == None)

        def show_leading_zeros(spin_button):
            adjustment = spin_button.get_adjustment()
            spin_button.set_text('{:02d}'.format(int(adjustment.get_value())))
            return True

        w("schedule-hours").connect("output", show_leading_zeros)
        w("schedule-minutes").connect("output", show_leading_zeros)

        w("backup-schedule-cancel").set_visible(True)
        if tooltip == None and self._build_exist_job():
            w("backup-schedule-cancel").set_visible(False)
            w("backup-schedule-delete").set_visible(True)

    def _init_backup_data(self):
        w = self.widget
        excluded = self.vm.get_xmlobj(inactive=True).get_backup_excluded()
        if excluded:
            backuper = self._build_backuper()
            for disk in backuper.get_diskinfos():
                if disk.is_clone_requested() and disk.disk.target in excluded:
                    disk.set_share_requested()

        folder = self.vm.get_xmlobj(inactive=True).get_backup_folder()
        if folder:
            new_folder = os.path.join(folder)
            w("backup-folder").set_text(new_folder)
        else:
            w("backup-folder").set_text(
                f"/backup/{self.vm.get_name()}")

        type = self.vm.get_xmlobj(inactive=True).get_backup_type()
        if type:
            index = self._backup_types.index(type.capitalize())
            w("backup-type").set_active(index)

        threshold = self.vm.get_xmlobj(inactive=True).get_backup_threshold()
        if threshold:
            w("threshold").set_value(int(threshold) / 1024 / 1024)
            w("threshold-label").set_sensitive(True)
            w("threshold").set_sensitive(True)
            w("threshold-postfix").set_sensitive(True)
            w("threshold-checkbox").set_active(True)

        compress = self.vm.get_xmlobj(inactive=True).get_backup_compress()
        if compress:
            w("compress-level").set_value(int(compress))
            w("compress-level-label").set_sensitive(True)
            w("compress-level").set_sensitive(True)
            w("compress-level").set_draw_value(True)
            w("compress-checkbox").set_active(True)

        repeat = self.vm.get_xmlobj(inactive=True).get_backup_repeat()
        schedule = self.vm.get_xmlobj(inactive=True).get_backup_schedule()
        if repeat and schedule:
            # Синхронизация из crontab
            job = self._build_exist_job()
            time = str(job).split("virtnbdbackup")[0].strip()

            if time != "None" and schedule != time:
                schedule = time
                self.vm.define_backup_schedule(schedule)

            # Обновление UI
            index = self._repeat_types.index(repeat.capitalize())
            w("schedule-repeat").set_active(index)

            if schedule[0] == "@":
                return
            minute, hour, day, month, daynum = (schedule.split(" "))

            if minute != "*":
                w("schedule-minutes").set_value(int(minute))
            if hour != "*":
                w("schedule-hours").set_value(int(hour))
            if day != "*":
                self.widget("schedule-calendar").select_day(int(day))

        w("backup-encryption-cancel").set_visible(True)
        encryption_key = self.vm.get_xmlobj(inactive=True).get_backup_encryption_key()
        if encryption_key:
            if os.path.exists(encryption_key):
                w("encryption-key-path").set_text(encryption_key)
                w("encryption-key-path").set_sensitive(True)
                w("label-encryption-key-path").set_sensitive(True)
                w("label-encryption-key").set_sensitive(False)
                w("encryption-key").set_sensitive(False)
                w("radio-key-path").set_active(True)
                uiutil.set_list_selection(w("encryption-key"), None)
            else:
                uiutil.set_list_selection(w("encryption-key"), encryption_key)

            w("backup-encryption-cancel").set_visible(False)
            w("backup-encryption-delete").set_visible(True)
            self._encryption_key = encryption_key
            w("encryption-icon").set_from_icon_name('dialog-ok',
                                                    Gtk.IconSize.BUTTON)
            w("compress-level-label").set_sensitive(True)
            w("compress-level").set_sensitive(True)
            w("compress-level").set_draw_value(True)
            w("compress-checkbox").set_active(True)
            w("compress-checkbox").set_sensitive(False)

    def _reset_state(self):
        self.widget("backup-cancel").grab_focus()

        text = self.widget("backup-folder").get_text()
        self.widget("backup-ok").set_sensitive(bool(text))

        # Populate default backup values
        backuper = self._build_backuper()
        self.widget("backup-orig-name").set_text(backuper.src_name)

        uiutil.set_grid_row_visible(
            self.widget("backup-dest-host"), self.conn.is_remote())
        self.widget("backup-dest-host").set_text(self.conn.get_pretty_desc())

        # Collect info about all the VMs disks
        self._storage_list = collections.OrderedDict()

        for diskinfo in backuper.get_diskinfos():
            sinfo = _StorageInfo(self.vm, backuper, diskinfo)
            if diskinfo.is_clone_requested():
                self._storage_list[sinfo.get_target()] = sinfo

        self._populate_storage_ui()

    def _cron_tooltip(self):
        def check_package(package_name):
            import subprocess
            try:
                subprocess.check_output(['dpkg-query', '-W', '-f=${Status}',
                                         package_name])
                return True
            except subprocess.CalledProcessError:
                return False

        tooltip = ""
        if not check_package("python3-crontab"):
            if len(tooltip):
                tooltip += "\n"
            tooltip += "Python3-crontab isn't installed"
        try:
            import os
            from crontab import CronTab
            CronTab(user=os.getlogin())
        except:
            if len(tooltip):
                tooltip += "\n"
            tooltip += "No rights for crontab"
        if not len(tooltip):
            return None
        return _(tooltip)


    #################
    # Value getters #
    #################

    def get_ui_backup_folder(self):
        return self.widget("backup-folder").get_text()

    def get_ui_backup_type(self):
        return uiutil.get_list_selection(self.widget("backup-type"))

    def get_ui_compress_active(self):
        return self.widget("compress-checkbox").get_active()

    def get_ui_compress_level(self):
        return int(self.widget("compress-level").get_value())

    def get_ui_threshold_active(self):
        return self.widget("threshold-checkbox").get_active()

    def get_ui_threshold(self):
        return int(self.widget("threshold").get_text())

    def get_ui_calendar(self):
        return self.widget("schedule-calendar").get_date()

    def get_ui_hours(self):
        return self.widget("schedule-hours").get_text()

    def get_ui_minutes(self):
        return self.widget("schedule-minutes").get_text()

    def get_ui_repeat(self):
        return uiutil.get_list_selection(self.widget("schedule-repeat"))

    def get_ui_encryption_key(self):
        if self.widget("radio-existing-key").get_active():
            return uiutil.get_list_selection(self.widget("encryption-key"))
        else:
            key_path = self.widget("encryption-key-path").get_text()
            gpg = gnupg.GPG()
            if not os.path.exists(key_path):
                self.err.show_err(_("The path does not exist"))
                return False

            with open(key_path, 'r') as key_file:
                key_data = key_file.read()

            import_result = gpg.import_keys(key_data)
            if import_result.count == 0:
                self.err.show_err(_("The file does not contain a valid GPG key."))
                return False
            return key_path
    
    def get_ui_hash(self):
        return uiutil.get_list_selection(self.widget("backup-hash"))


    #######################
    # Functional routines #
    #######################

    def _build_backuper(self):
        conn = self.conn.get_backend()
        orig_name = self.vm.get_name()
        backuper = Backuper(conn, src_name=orig_name)
        return backuper

    def _build_exist_job(self):
        cron = CronTab(user=os.getlogin())
        iter = cron.find_comment(self.vm.get_cron_comment())
        job = next(iter, None)
        return job


    ###############
    # UI building #
    ###############

    @property
    def _backup_types(self):
        return sorted(["Auto", "Full", "Inc", "Diff", "Copy"])

    @property
    def _repeat_types(self):
        return sorted(
            [_("Hourly"), _("Daily"), _("Weekly"), _("Monthly"),
             _("Every 3th month"), _("Every 6th month")])

    def _populate_storage_ui(self):
        """
        Fill in the storage UI from the collected StorageInfo classes
        """
        model = self.widget("storage-list").get_model()
        model.clear()
        excluded = self.vm.get_xmlobj(inactive=True).get_backup_excluded()
        for sinfo in self._storage_list.values():
            if (excluded and sinfo.get_target() in excluded):
                sinfo.set_clone_requested(False)
            model.append([sinfo.get_target(), sinfo.get_tooltip()])
            model.append(["separator", None])

    def _storage_backup_toggled_cb(self, src, treepath):
        model = self.widget("storage-list").get_model()
        row = model[treepath]
        target = row[0]
        sinfo = self._storage_list[target]
        active = not src.get_property("active")
        sinfo.set_clone_requested(active)
        model.emit("row-changed", row.path, row.iter)

    def _populate_backup_type_combo(self, combo):
        uiutil.build_simple_combo(
            combo, [[type.lower(), type] for type in self._backup_types])

    def _populate_repeat_combo(self, combo):
        uiutil.build_simple_combo(
            combo, [[type.lower(), type] for type in self._repeat_types])
        index = self._repeat_types.index(_("Hourly"))
        combo.set_active(index)

    def _populate_encryption_key_combo(self, combo):
        gpg = gnupg.GPG()
        keys = gpg.list_keys()

        rows = []
        for key in keys:
            time = datetime.datetime.fromtimestamp(int(key['date'])).strftime('%Y-%m-%d %H:%M:%S')
            rows.append([key['fingerprint'], f"{time}: {key['keyid']}"])
        uiutil.build_simple_combo(combo, rows)

    def _populate_hash_combo(self, combo):
        hashes = ["adler32", "xxh64"]
        uiutil.build_simple_combo(
            combo, [[hash, hash] for hash in hashes])
        combo.set_active(0)


    ##########################
    # browse folder handling #
    ##########################

    def _browse_folder(self):
        filepath = self.err.browse_local(dialog_name=_("Select folder"),
            dialog_type=Gtk.FileChooserAction.SELECT_FOLDER)
        if filepath:
            self.widget("backup-folder").set_text(filepath)
            self.widget("backup-ok").set_sensitive(True)

    def _browse_folder_changed(self):
        text = self.widget("backup-folder").get_text()
        self.widget("backup-ok").set_sensitive(bool(text))


    #################
    # cron handling #
    #################

    def _build_cron(self, command):
        cron = CronTab(user=os.getlogin())
        job = cron.new(command=command, comment=self.vm.get_cron_comment())

        year, month, day = self.widget("schedule-calendar").get_date()
        hour = self.get_ui_hours()
        minute = self.get_ui_minutes()
        repeat = self.get_ui_repeat().capitalize()

        if repeat == _("Hourly"):
            job.minute.on(minute)
        if repeat == _("Daily"):
            job.minute.on(minute)
            job.hour.on(hour)
        if repeat == _("Weekly"):
            daynum = (datetime.date(year, month + 1, day).weekday() + 1) % 7
            job.minute.on(minute)
            job.hour.on(hour)
            job.dow.on(daynum)
        if repeat == _("Monthly"):
            job.minute.on(minute)
            job.hour.on(hour)
            job.day.on(day)
        if repeat == _("Every 3th month"):
            job.minute.on(minute)
            job.hour.on(hour)
            job.day.on(day)
            job.month.on(3)
        if repeat == _("Every 6th month"):
            job.minute.on(minute)
            job.hour.on(hour)
            job.day.on(day)
            job.month.on(6)
        return (cron, job)

    def _build_command(self, backuper):
        vm_name = self.vm.get_name()

        command = (f"virtnbdbackup -S -n -U {self.conn.get_uri()} -d {vm_name} "
                   f"-l {self.get_ui_backup_type()} "
                   f"-o {self.get_ui_backup_folder()} "
                   f"--scratchdir /run/libvirt")
        
        if self.get_ui_threshold_active():
            command += f" --threshold {self.get_ui_threshold() * 1024 * 1024}"
        
        if self.get_ui_compress_active():
            command += f" --compress {self.get_ui_compress_level()}"

        counter = 0
        has_row_disk = False
        for disk in backuper.get_diskinfos():
            if disk.disk.driver_type == "raw" and disk.is_clone_requested():
                has_row_disk = True

            if not disk.get_share_msg() and disk.is_share_requested():
                if counter == 0:
                    command += f" -x {disk.disk.target}"
                else:
                    command += f",{disk.disk.target}"
                counter += 1
        
        if has_row_disk:
            command += " --raw"

        if self._encryption_key and os.path.exists(self._encryption_key):
            command += f" --encryption-key-path {self._encryption_key}"

            gpg = gnupg.GPG()
            keys = gpg.list_keys()

            with open(self._encryption_key, 'r') as key_file:
                key_data = key_file.read()

            fingerprint = gpg.import_keys(key_data).results[0]['fingerprint']
            for key in keys:
                if key['fingerprint'] == fingerprint:
                    command += f" --encryption-key-fingerprint {key['fingerprint']}"
                    break
        elif self._encryption_key:
            command += f" --encryption-key-fingerprint {self._encryption_key}"
        
        command += f" --hash {self.get_ui_hash()}"

        return command


    ##############################
    # Subdialog windows handling #
    ##############################

    def _schedule_dialog_finish(self, src):
        self._schedule_dialog_close()

        backuper = self._build_final_backuper()
        command = self._build_command(backuper)
        cron, new_job = self._build_cron(command)
        jobs = list(cron.find_comment(new_job.comment))
        jobs_set = set(cron.find_comment(new_job.comment))
        if len(jobs_set) > 1:
            res = self.err.yes_no(
                _("Are you sure you want to change backup job?"),
                _("Old backup job will be removed."))
            if not res:
                return
            for job in jobs_set:
                if job != new_job:
                    cron.remove(job)
        if len(jobs_set) == 1 and len(jobs) > 1:
            self.close()

        cron.write()
        self._update_guest_config(backuper, new_job)
        self.err.show_info(_("The job has been created"), str(new_job))
        self.close()

    def _show_schedule_window(self, src):
        self.widget("vmm-backup-schedule").show_all()

    def _schedule_delete(self):
        try:
            self.vm.undefine_backup_schedule()
            cron = CronTab(user=os.getlogin())
            iter = cron.find_comment(self.vm.get_cron_comment())
            job = next(iter, None)
            if job:
                cron.remove_all(comment=job.comment)
                cron.write()
                self.widget("backup-schedule-delete").set_visible(False)
                self.widget("backup-schedule-cancel").set_visible(True)
        except Exception as e:
            log.debug("Exception in schedule_delete: %s", e)

    def _encryption_dialog_finish(self, src):
        w = self.widget
        key = self.get_ui_encryption_key()
        if key:
            self._encryption_key = key
            w("encryption-icon").set_from_icon_name('dialog-ok',
                                                    Gtk.IconSize.BUTTON)
            w("compress-level-label").set_sensitive(True)
            w("compress-level").set_sensitive(True)
            w("compress-level").set_draw_value(True)
            w("compress-checkbox").set_active(True)
            w("compress-checkbox").set_sensitive(False)
            w("backup-encryption-delete").set_visible(True)
            w("backup-encryption-cancel").set_visible(False)
        self._encryption_dialog_close()

    def _show_encryption_window(self, src):
        self.widget("vmm-backup-encryption").show_all()
    
    def _encryption_delete(self):
        w = self.widget
        self._encryption_key = None

        w("backup-encryption-delete").set_visible(False)
        w("backup-encryption-cancel").set_visible(True)
        w("encryption-icon").set_from_icon_name('dialog-error',
                                                Gtk.IconSize.BUTTON)
        compress = self.vm.get_xmlobj(inactive=True).get_backup_compress()
        w("compress-level").set_value(int(compress) if compress else 1)
        w("compress-level-label").set_sensitive(bool(compress))
        w("compress-level").set_sensitive(bool(compress))
        w("compress-level").set_draw_value(bool(compress))
        w("compress-checkbox").set_active(bool(compress))
        w("compress-checkbox").set_sensitive(True)

        self._encryption_dialog_close()


    ###################
    # Finish routines #
    ###################

    def _update_guest_config(self, backuper, job, without_schedule=False):
        excluded = included = ""
        for disk in backuper.get_diskinfos():
            if not disk.get_share_msg() and disk.is_share_requested():
                excluded += f" {disk.disk.target}"
            elif not disk.get_share_msg():
                included += f" {disk.disk.target}"

        threshold = (self.get_ui_threshold() * 1024 * 1024
                     if self.get_ui_threshold_active()
                     else None)
        compression = (self.get_ui_compress_level()
                       if self.get_ui_compress_active()
                       else None)
        time = repeat = None
        if job:
            time = str(job).split("virtnbdbackup")[0].strip()
            repeat = self.get_ui_repeat()

        self.vm.define_backup(time, repeat,
            self.get_ui_backup_folder(), self.get_ui_backup_type(),
            threshold, excluded.strip() or None, compression,
            included.strip() or None, self._encryption_key, without_schedule,
            self.get_ui_hash())

    def _finish_cb(self, error, details, conn):
        self.reset_finish_cursor()

        if error is not None:
            error = (_("Error during virtual machine backup '%(vm)s': "
                       "%(error)s") % {
                     "vm": self.vm.get_name(),
                     "error": error,
                     })
            self.err.show_err(error, details=details)
            return

        conn.schedule_priority_tick(pollvm=True)
        self.vm.emit("state-changed")
        self.close()

    def _async_backup(self, asyncjob, backuper, command):
        meter = asyncjob.get_meter()
        backuper.start_backuping(command, meter)

    def _build_final_backuper(self):
        backuper = self._build_backuper()
        for diskinfo in backuper.get_diskinfos():
            if diskinfo.is_clone_requested():
                target = diskinfo.disk.target
                sinfo = self._storage_list[target]
                sinfo.set_values_on_diskinfo(diskinfo)

        return backuper

    def _finish(self):
        backuper = self._build_final_backuper()
        command = self._build_command(backuper)

        title = (_("Creating virtual machine backup '%s'") % self.vm.get_name())
        text = ""
        if backuper.get_nonshare_diskinfos():
            text = (_("Downloading selected storages and creating "
                      "xml files (this may take a while)"))

        self.set_finish_cursor()
        self._update_guest_config(backuper, None, without_schedule=True)
        progWin = vmmAsyncJob(self._async_backup, [backuper, command],
                              self._finish_cb, [self.conn], title, text,
                              self.topwin)
        progWin.run()

    def _clear_folder(self, folder_path):
        try:
            for filename in os.listdir(folder_path):
                file_path = os.path.join(folder_path, filename)
                try:
                    if os.path.isfile(file_path) or os.path.islink(file_path):
                        os.unlink(file_path)
                        log.debug(f"File deleted: {file_path}")
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)
                        log.debug(f"Deleted folder: {file_path}")
                except Exception as e:
                    log.debug(f"File was not deleted: {e}")
        except Exception as e:
            log.debug(f"Failed to clear folder contents: {e}")


    #################
    # Delete backup #
    #################

    def _backup_delete(self):
        vm_status = self.vm.status()
        if self.vm.is_shutoff():
            self.vm.startup()
            self.vm.suspend()
        
        backup_folder = self.vm.get_xmlobj(inactive=True).get_backup_folder()
        self.vm.undefine_backup()
        if os.path.exists(backup_folder):
            self._clear_folder(backup_folder)

        checkpoints = self.get_checkpoints()
        self.delete_checkpoints(checkpoints)
        
        self.widget("backup-delete").set_visible(False)
        self.widget("backup-cancel").set_visible(True)

        if vm_status == libvirt.VIR_DOMAIN_SHUTOFF:
            self.vm.resume()
            self.vm.destroy()

    def get_checkpoints(self):
        command = ['virsh', '-c', self.conn.get_uri(), 'checkpoint-list',
                   self.vm.get_name()]

        try:
            result = subprocess.run(command, capture_output=True, text=True, check=True)
            checkpoints = result.stdout.strip().split('\n')
            if len(checkpoints) > 2:
                checkpoints = checkpoints[2:]
            checkpoint_list = [line.split()[0] for line in checkpoints if line and 'virtnbdbackup' in line]
            return checkpoint_list

        except subprocess.CalledProcessError as e:
            log.debug(f"Error executing command: {e}")
            return []

    def delete_checkpoints(self, checkpoints):
        for checkpoint in checkpoints:
            command = ['virsh', '-c', self.conn.get_uri(), 'checkpoint-delete',
                       self.vm.get_name(), checkpoint]

            try:
                result = str(subprocess.run(command, check=True))
                log.debug(result)
                log.debug(f"Deleted checkpoint: {checkpoint}")
            except subprocess.CalledProcessError as e:
                log.debug(f"Error deleting checkpoint {checkpoint}: {e}")


    ################
    # UI listeners #
    ################

    def _close_cb(self, src, event=None):
        return self.close()
    def _finish_clicked_cb(self, src):
        return self._finish()

    def _browse_local_clicked_cb(self, src):
        ingore = src
        self._browse_folder()

    def _backup_folder_changed_cb(self, src):
        ingore = src
        self._browse_folder_changed()

    def _backup_delete_cb(self, src):
        res = self.err.yes_no(
            _("Are you sure you want to delete backup folder and job?"))
        if not res:
            return
        self._backup_delete()
        self._schedule_delete()
        self.close()

    def _schedule_dialog_close_cb(self, src, event=None):
        return self._schedule_dialog_close()

    def _schedule_delete_cb(self, src):
        self._schedule_delete()

    def _schedule_dialog_finish_cb(self, src):
        self._schedule_dialog_finish(src)

    def _schedule_backup_cb(self, src):
        self._show_schedule_window(src)

    def _encryption_dialog_close_cb(self, src, event=None):
        return self._encryption_dialog_close()

    def _encryption_delete_cb(self, src):
        self._encryption_delete()

    def _encryption_dialog_finish_cb(self, src):
        self._encryption_dialog_finish(src)

    def _backup_encrypt_cb(self, src):
        self._show_encryption_window(src)
    
    def _encryption_key_toggled_cb(self, src):
        active = bool(src.get_active())
        self.widget("encryption-key").set_sensitive(active)
        self.widget("label-encryption-key").set_sensitive(active)
        self.widget("encryption-key-path").set_sensitive(not active)
        self.widget("label-encryption-key-path").set_sensitive(not active)
    
    def _threshold_toggled_cb(self, src):
        active = src.get_active()
        self.widget("threshold-label").set_sensitive(active)
        self.widget("threshold").set_sensitive(active)
        self.widget("threshold-postfix").set_sensitive(active)

    def _compression_toggled_cb(self, src):
        active = src.get_active()
        self.widget("compress-level-label").set_sensitive(active)
        self.widget("compress-level").set_sensitive(active)
        self.widget("compress-level").set_draw_value(active)
