import re
import libvirt
from virtinst import log

from .lib import uiutil
from .baseclass import vmmGObjectUI


GROUP_RENAME = "rename"

class vmmGroupDialog(vmmGObjectUI):
    _instance = None

    @classmethod
    def reset_instance(cls):
        cls._instance = None

    @classmethod
    def show_instance(cls, parentobj, vm, type=None):
        try:
            if cls._instance:
                cls._instance.close()
            cls.reset_instance()
            cls._instance = vmmGroupDialog(vm, type)
            cls._instance.show(parentobj.topwin, vm)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(
                    _("Error launching group dialog: %s") % str(e))

    def __init__(self, vm, type):
        vmmGObjectUI.__init__(self, "group.ui", "vmm-group")
        self.vm = vm
        self.type = type

        self.builder.connect_signals({
            "on_vmm_group_delete_event": self._delete_event,
            "on_group_cancel_clicked": self._cancel_clicked,
            "on_group_finish_clicked": self._finish_clicked,
        })

        self._build_group_combo(self.vm, self.widget("group-model"))

    def _cleanup(self):
        self.vm = None
        self.type = None

    @property
    def conn(self):
        if isinstance(self.vm, list):
            return self.vm[0] and self.vm[0].conn or None
        return self.vm and self.vm.conn or None

    @property
    def groups(self):
        groups = []
        for vm in self.conn.list_vms():
            group = vm.get_xmlobj(inactive=True).get_group()
            if group and group not in groups:
                groups.append(group)
        return groups


    ##############
    # Public API #
    ##############

    def show(self, parent, vm):
        log.debug("Showing group wizard")
        self._build_group_combo(self.vm, self.widget("group-model"))
        self._set_vm(vm)
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing group wizard")
        self.topwin.hide()
        return 1

    def _vm_removed_cb(self, _conn, vm):
        if self.vm == vm:
            self.close()

    def _set_vm(self, newvm):
        oldvm = self.vm
        if oldvm:
            if isinstance(oldvm, list):
                for vm in oldvm:
                    vm.conn.disconnect_by_obj(self)
            else:
                oldvm.conn.disconnect_by_obj(self)
        if newvm:
            if isinstance(newvm, list):
                for vm in newvm:
                    vm.conn.connect("vm-removed", self._vm_removed_cb)
            else:
                newvm.conn.connect("vm-removed", self._vm_removed_cb)
        self.vm = newvm


    #############
    # Listeners #
    #############

    def _delete_event(self, ignore1, ignore2):
        self.close()
        return 1

    def _cancel_clicked(self, src):
        ignore = src
        self.close()

    def _finish_clicked(self, src):
        ignore = src
        self._finish()


    ##################
    # group handling #
    ##################

    def _finish(self):
        group_name = self.widget("group-model-text").get_text()
        if not group_name:
            self._handle_delete()
            return

        valid_name = self._handle_name_validation(group_name)
        if not valid_name:
            self.close()
            return

        if self.type == GROUP_RENAME:
            self._handle_rename(group_name)
        else:
            if isinstance(self.vm, list):
                for vmm in self.vm:
                    vmm.define_group(group_name)
            else:
                self.vm.define_group(group_name)
        self.close()

    def _handle_name_validation(self, group_name):
        regex = re.compile('[&<>\"\']')
        if regex.search(group_name):
            self.err.show_err(_("Group name contains invalid characters (&, <, >, ', \")."))
            return False

        if group_name.isspace():
            self.err.show_err(_("Group name cannot consist of only spaces."))
            return False
        return True

    def _handle_delete(self):
        old_group = ""
        if isinstance(self.vm, list):
            for vmm in self.vm:
                old_group += f"{vmm.get_xmlobj(inactive=True).get_group()}, "
            old_group = old_group[:-2]
        else:
            old_group = self.vm.get_xmlobj(inactive=True).get_group()
        if old_group:
            result = self.err.yes_no(
                _("This will remove the group:\n\n%s\n\n"
                    "Are you sure?") % old_group)
            if result:
                if isinstance(self.vm, list):
                    for vmm in self.vm:
                        vmm.delete_vm_group()
                else:
                    self.vm.delete_vm_group()
        self.close()
    
    def _handle_rename(self, group_name):
        if group_name in self.groups:
            self.err.show_err(_("A group with this name already exist"))
            return
        if isinstance(self.vm, list):
            for vmm in self.vm:
                vmm.rename_group(group_name)
        else:
            self.vm.rename_group(group_name)

    def _build_group_combo(self, vm, combo):
        uiutil.build_simple_combo(combo, 
                                  [[group, group] for group in self.groups])
        if isinstance(vm, list):
            return
        group_name = vm.get_xmlobj(inactive=True).get_group()
        if group_name:
            index = sorted(self.groups).index(group_name)
            combo.set_active(index)
