import datetime
import glob
import os
import re
import subprocess
import shutil

from virtManager.asyncjob import vmmAsyncJob
from virtManager.lib import uiutil
from virtManager.baseclass import vmmGObjectUI
from virtinst import log, xmlutil
from virtinst.guest import Guest
from virtinst.importer import _upload_disk
from virtinst.storage import StoragePool, StorageVolume

from gi.repository import Gtk, Gdk
from gi.repository import Pango


class _BackupInfo:
    """
    Wrapper class for backup row
    """
    def __init__(self, conn, path, checkpoints, files, pool_name, vm=None):
        self.conn = conn
        self._path = path
        self._checkpoints = checkpoints
        self._files = files
        self._pool_name = pool_name
        self._vm = vm

        self._checkpoint_name = checkpoints[-1] # last one

        self._is_restore_requested = False
        self._guest = None

        self._generate_guest()

    def _generate_guest(self):
        last_config = glob.glob(os.path.join(self._path, "vmconfig*.xml"))[-1]
        with open(last_config, "r") as f:
            src_xml = f.read()
            self._guest = Guest(self.conn, parsexml=src_xml)

    def get_guest_name(self):
        return self._guest.name
    def set_pool_name(self, pool_name):
        self._pool_name = pool_name
    def get_pool_name(self):
        return self._pool_name
    def get_checkpoint_name(self):
        return self._checkpoint_name
    def set_checkpoint_name(self, checkpoint_tuple):
        self._checkpoint_name = checkpoint_tuple

    def get_path(self):
        return self._path
    def get_checkpoints(self):
        return self._checkpoints
    def get_files(self):
        return self._files
    def get_vm(self):
        return self._vm

    def is_restore_requested(self):
        return self._is_restore_requested
    def set_restore_requested(self, val):
        self._is_restore_requested = bool(val)

    def get_tooltip(self):
        vm = self.get_vm()
        if vm:
            return (_("This folder is used by the virtual machine '%s'")
                    % vm.get_name())
    def get_markup(self):
        lines = []

        lines.append(_("Path: %s") % self.get_path())
        lines.append(_("VM name: %s") % self.get_guest_name())

        label = "\n".join(lines)
        markup = "<small>%s</small>" % xmlutil.xml_escape(label)
        return markup
    def get_icon(self):
        if self.get_vm():
            return "folder-saved"


class vmmRestoreDialog(vmmGObjectUI):
    _instance = None

    @classmethod
    def reset_instance(cls):
        cls._instance = None

    @classmethod
    def show_instance(cls, parentobj, conn):
        try:
            if cls._instance:
                cls._instance.close()
            cls.reset_instance()
            cls._instance = vmmRestoreDialog(conn)
            cls._instance.show(parentobj.topwin)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(
                _("Error launching restore dialog: %s") % str(e))

    def __init__(self, conn):
        vmmGObjectUI.__init__(self, "restore.ui", "vmm-restore")
        self.conn = conn
        self._backups = {}

        self.builder.connect_signals({
            "on_restore_delete_event": self._close_cb,
            "on_restore_cancel_clicked": self._close_cb,
            "on_restore_ok_clicked": self._finish_clicked_cb,

            "on_backup_folder_changed": self._backup_folder_changed_cb,
            "on_browse_backup_local_clicked": self._browse_backup_local_clicked_cb,
            "on_browse_restore_local_clicked": self._browse_restore_local_clicked_cb,
            "on_action_type_changed": self._action_type_changed_cb,
        })

        self._init_ui()

    def _cleanup(self):
        self.conn = None
        self._backups = None

    def _init_ui(self):
        backup_list = self.widget("backup-folders")

        # [path, tooltip, pool, checkpoint]
        self.model = Gtk.ListStore(str, str, str, str)
        backup_list.set_model(self.model)
        backup_list.set_tooltip_column(1)

        backupCol = Gtk.TreeViewColumn(_("Select all"))
        backupCol.set_clickable(True)
        backupCol.connect("clicked", self.select_all)
        pathCol = Gtk.TreeViewColumn(_("Backup"))
        pathCol.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        pathCol.set_expand(True)
        poolCol = Gtk.TreeViewColumn(_("Pool"))
        checkpointCol = Gtk.TreeViewColumn(_("Checkpoint"))

        backup_list.append_column(backupCol)
        backup_list.append_column(pathCol)
        backup_list.append_column(poolCol)
        backup_list.append_column(checkpointCol)

        def separator_cb(_model, _iter):
            return _model[_iter][0] == "separator"
        backup_list.set_row_separator_func(separator_cb)

        chkbox = Gtk.CellRendererToggle()
        chkbox.connect('toggled', self._backup_toggled_cb)
        backupCol.pack_start(chkbox, False)

        def chk_cell_data_cb(column, cell, model, _iter, data):
            target = model[_iter][0]
            if target == "separator":
                return
            backup = self._backups[target]
            active = backup.is_restore_requested()
            # sensitive = model[_iter][2] != _("Select pool")
            # _chkbox.set_property('sensitive', sensitive)
            _chkbox = column.get_cells()[0]
            _chkbox.set_property('active', active)
            tooltip = backup.get_tooltip()
            if tooltip != model[_iter][1]:
                model[_iter][1] = tooltip

        backupCol.set_cell_data_func(chkbox, chk_cell_data_cb)

        pathtxt = Gtk.CellRendererText()
        pathtxt.set_property("width-chars", 30)
        pathtxt.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)
        pathimg = Gtk.CellRendererPixbuf()
        pathimg.set_property('stock-size', Gtk.IconSize.MENU)
        pathimg.set_padding(3, 0)
        pool_model = Gtk.ListStore(str)
        pool_combo = Gtk.CellRendererCombo()
        pool_combo.set_property("editable", True)
        pool_combo.set_property("model", pool_model)
        pool_combo.set_property("text-column", 0)
        pool_combo.set_property("has-entry", False)
        pool_combo.connect("edited", self.on_pool_combo_edited)
        pool_combo.set_padding(5, 0)
        checkpoint_model = Gtk.ListStore(str, str)
        checkpoint_combo = Gtk.CellRendererCombo()
        checkpoint_combo.set_property("editable", True)
        checkpoint_combo.set_property("model", checkpoint_model)
        checkpoint_combo.set_property("text-column", 0)
        checkpoint_combo.set_property("has-entry", False)
        checkpoint_combo.connect("edited", self.on_checkpoint_combo_edited)
        checkpoint_combo.connect("editing-started", self.on_checkpoint_combo_clicked)
        pathCol.pack_start(pathimg, False)
        pathCol.pack_start(pathtxt, True)
        poolCol.pack_start(pool_combo, True)
        checkpointCol.pack_start(checkpoint_combo, True)

        for pool in self.conn.list_pools():
            pool_model.append([pool.get_name()])

        def path_cb(column, cell, model, _iter, data):
            target = model[_iter][0]
            if target == "separator":
                return
            backup = self._backups[target]
            _pathimg = column.get_cells()[0]
            _pathtxt = column.get_cells()[1]
            _pathimg.set_property('icon-name', backup.get_icon())
            _pathtxt.set_property('markup', backup.get_markup())
        
        def pool_cb(column, cell, model, _iter, data):
            target = model[_iter][0]
            if target == "separator":
                return
            backup = self._backups[target]
            _pool_combo = column.get_cells()[0]
            _pool_combo.set_property('text', backup.get_pool_name())

        def checkpoint_cb(column, cell, model, _iter, data):
            target = model[_iter][0]
            if target == "separator":
                return
            backup = self._backups[target]
            _checkpoint_combo = column.get_cells()[0]
            _checkpoint_combo.set_property('text', backup.get_checkpoint_name()[0])

        pathCol.set_cell_data_func(pathtxt, path_cb)
        pathCol.set_cell_data_func(pathimg, path_cb)
        poolCol.set_cell_data_func(pool_combo, pool_cb)
        checkpointCol.set_cell_data_func(checkpoint_combo, checkpoint_cb)

        self.widget("backup-dest-host").set_text(self.conn.get_uri())
        default_backuppath = self.config.get_default_directory("backup")
        if default_backuppath:
            self.widget("backup-folder").set_text(default_backuppath)
        self._populate_action_type_combo(self.widget("action-type"))
        self.widget("restore-ok").set_sensitive(False)


    ################
    # UI callbacks #
    ################

    def on_pool_combo_edited(self, widget, path, new_text):
        target = self.model[path][0]
        if target == "separator":
            return
        backup = self._backups[target]
        backup.set_pool_name(new_text)
        self.model[path][2] = new_text

    def on_checkpoint_combo_edited(self, widget, path, new_text):
        target = self.model[path][0]
        if target == "separator":
            return
        backup = self._backups[target]
        model = widget.get_property('model')
        _iter = model.get_iter_first()
        checkpoint = (model[_iter][0], model[_iter][1])
        while _iter:
            if model[_iter][0] == new_text:
                checkpoint = (model[_iter][0], model[_iter][1])
                break
            _iter = model.iter_next(_iter)
        backup.set_checkpoint_name(checkpoint)
        self.model[path][3] = new_text

    def on_checkpoint_combo_clicked(self, widget, ignore, path):
        target = self.model[path][0]
        if target == "separator":
            return
        backup = self._backups[target]
        model = widget.get_property('model')
        model.clear()
        for checkpoint in backup.get_checkpoints():
            model.append([checkpoint[0], checkpoint[1]])
    
    def select_all(self, col):
        counter = 0
        for backup in self._backups.values():
            if (col.get_title() == _("Select all")
                and backup.get_pool_name() != _("Select pool")):
                    backup.set_restore_requested(True)
                    counter += 1
            if (col.get_title() == _("Unselect all")
                and backup.get_pool_name() != _("Select pool")):
                    backup.set_restore_requested(False)
                    counter += 1
        if counter:
            if col.get_title() == _("Select all"):
                col.set_title(_("Unselect all"))
            else:
                col.set_title(_("Select all"))

    def _backup_toggled_cb(self, src, treepath):
        model = self.widget("backup-folders").get_model()
        row = model[treepath]
        target = row[0]
        backup = self._backups[target]
        active = not src.get_property("active")
        backup.set_restore_requested(active)
        model.emit("row-changed", row.path, row.iter)

        checked = [backup for backup in self._backups.values()
                   if backup.is_restore_requested()]
        self.widget("restore-ok").set_sensitive(checked)


    ##############
    # Public API #
    ##############

    def show(self, parent):
        log.debug("Showing restore wizard")
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing restore wizard")
        self.topwin.hide()
        return 1


    ##############
    # UI helpers #
    ##############

    def _browse_backup_folder(self):
        folderpath = self.err.browse_local(dialog_name=_("Select folder"),
            dialog_type=Gtk.FileChooserAction.SELECT_FOLDER)
        if folderpath:
            self.widget("backup-folder").set_text(folderpath)
            self._load_backup_folders(folderpath)
            self.config.set_default_directory("backup", folderpath)

    def _browse_restore_folder(self):
        folderpath = self.err.browse_local(dialog_name=_("Select folder"),
            dialog_type=Gtk.FileChooserAction.SELECT_FOLDER)
        if folderpath:
            self.widget("restore-folder").set_text(folderpath)

    def _load_backup_folders(self, backuppath):
        exist_vms = {vm.get_xmlobj().get_backup_folder(): vm
                     for vm in self.conn.list_vms()}
        self._backups = {}
        backup_folders = next(os.walk(backuppath))[1]
        pool_name = self.conn.list_pools()[0].get_name()

        for backup_folder in backup_folders:
            try: 
                backup_files = next(os.walk(os.path.join(backuppath,
                                                         backup_folder)))
                checkpoint_files = next(os.walk(os.path.join(backup_files[0],
                                                             "checkpoints")))
                path = backup_files[0]
                counter = 1
                checkpoints = []
                for file in sorted(checkpoint_files[2], key=lambda x: (len(x), x)):
                    checkpoint_path = os.path.join(checkpoint_files[0], file)
                    m_time = os.path.getmtime(checkpoint_path)
                    dt_m = datetime.datetime.fromtimestamp(m_time)
                    view = f"{counter}: " + str(dt_m).rsplit(".", 1)[0]
                    checkpoints.append((view, file))
                    counter += 1
                files = backup_files[2]

                vm = exist_vms[path] if path in exist_vms.keys() else None
                backup = _BackupInfo(self.conn, path, checkpoints, files, pool_name, vm)
                self._backups[backup_files[0]] = backup
            except:
                log.debug(f"Backup folder: {backup_folder} is broken.")

        self._populate_backups_ui()

    def _populate_backups_ui(self):
        model = self.widget("backup-folders").get_model()
        model.clear()
        for backup in self._backups.values():
            model.append([backup.get_path(), backup.get_tooltip(),
                          backup.get_pool_name(),
                          backup.get_checkpoint_name()[0]])
            model.append(["separator", None, None, None])

    def _populate_action_type_combo(self, combo):
        action_types = [_("_Delete"), _("Restore"),
                        _("Compare checksums")]
        uiutil.build_simple_combo(
            combo, [[type, type] for type in action_types])

    def _action_type_changed(self, action_type):
        self.widget("restore-folder-box").set_visible(
            action_type == _("Restore into folder"))


    ###################
    # Finish handlers #
    ###################

    def _finish(self):
        vms = [backup for backup in self._backups.values()
               if backup.is_restore_requested()]
        action_type = uiutil.get_list_selection(self.widget("action-type"))
        if action_type == _("_Delete"):
            res = self.err.yes_no(_("Are you sure you want to permanently "
                                    "delete the selected backups?"))
            if res:
                self._delete(vms)
        elif action_type == _("Restore"):
            self._restore(vms)
        else:
            self._compares_checksum(vms)

    def _delete(self, vms):
        def _clear_folder(folder_path):
            for filename in os.listdir(folder_path):
                file_path = os.path.join(folder_path, filename)
                try:
                    if os.path.isfile(file_path) or os.path.islink(file_path):
                        os.unlink(file_path)
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)
                except Exception as e:
                    log.debug(f"File was not deleted: {e}")

        for vm in vms:
            _clear_folder(vm.get_path())

        # Refresh
        folderpath = self.widget("backup-folder").get_text()
        self._load_backup_folders(folderpath)

    def _compares_checksum(self, vms):
        error_messages = []
        commands = []
        for vm in vms:
            commands.append((vm.get_path(),
                             f"virtnbdrestore -i {vm.get_path()} -o verify"))

        def _async_verify(commands, error_messages):
            for cmd in commands:
                result = subprocess.run(cmd[1], shell=True, capture_output=True,
                                        text=True)
                messages = f"\n{result.stderr}"
                log.debug(messages)

                messages = messages.split("\n")
                for message in messages:
                    if "ERROR" in message:
                        message = re.sub(
                            r'\x1b\[\d+m|\[\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\]',
                            '', message)
                        error_messages.append(
                            (cmd[0], message.split(":", 1)[1].strip()))

        def finish_cb():
            nonlocal error_messages
            def error_view(error):
                parts = error[1].split(":")
                message = _(parts[0])
                return (_("%(message)s in '%(vm)s':%(comparement)s") %
                        {"message": message, "vm": error[0],
                         "comparement": parts[1]})

            if error_messages:
                error_messages = [error_view(e) for e in error_messages]
                self.err.show_err(error_messages[0], "\n".join(error_messages))
            else:
                self.err.show_info(_("All checksums match"))

        title = _("Verification")
        text = _("Creation and verification of checksums")
        vmmAsyncJob.simple_async(_async_verify, [commands, error_messages],
                                 self, title, text, "", finish_cb=finish_cb)

    def _restore(self, vms):
        # Check for vm existance
        restored_vms = set([vm.get_guest_name() for vm in vms])
        existed_vms = set([vm.get_name() for vm in self.conn.list_vms()])
        interception = restored_vms & existed_vms
        if interception:
            self.err.show_err(_("To begin restore, delete existing virtual "
                                "machines: '%s'") % ", ".join(interception))
            return

        error_messages = []
        def _finish_cb(error, details):
            nonlocal error_messages
            if error_messages:
                error = (_("Error during virtual machine(s) restore."))
                error_messages = [err[0] + ": " + err[1] for err in error_messages]
                self.err.show_err(error, details="\n".join(error_messages))

        progWin = vmmAsyncJob(self._async_restore, [vms, error_messages],
                              _finish_cb, [], _("Restoring virtual machine(s)"),
                              _("This may take a while"), self.topwin)
        progWin.run()

    def _async_restore(self, asyncjob, vms, error_messages):
        for vm in vms:
            dom, disks = None, []
            # Build command
            output = os.path.join("/tmp/restore", vm.get_guest_name())
            command = (f"virtnbdrestore -n -U {self.conn.get_uri()}"
                       f" -o {output} -i {vm.get_path()} --raw"
                       f" --until {vm.get_checkpoint_name()[1]}")

            try:
                self.create_tmp_folder(output)
                self.subprocess_restore(command, error_messages,
                                        vm.get_guest_name())
                # Delete restored folder if error
                if error_messages and error_messages[-1][0] == vm.get_guest_name():
                    raise

                # Generate guest
                guest = self.generate_guest(output)

                # Upload disks
                disk_names = list(filter(lambda file: not file.endswith('.xml'),
                                         os.listdir(output)))
                existed_disks = self.existed_disks(disk_names[:],
                                                   vm.get_pool_name())
                log.debug(existed_disks)
                if existed_disks:
                    text = ("Storage(s) with name '%s' already exist") % existed_disks
                    error_messages.append((vm.get_guest_name(), text))
                    raise Exception(text)

                conn = self.conn.get_backend()
                for disk_name in disk_names:
                    self.save_disk(output, disk_name, conn, vm.get_pool_name(),
                                   guest)
                    disks.append(disk_name)
                dom = conn.defineXML(guest.get_xml())

            except Exception as e:
                log.debug("Restoring failed: %s", str(e))
                if dom:
                    dom.undefine()
                if disks:
                    for disk in disks:
                        conn = self.conn.get_backend()
                        pool = StoragePool(conn, pool.XMLDesc(0))
                        path = os.path.join(pool.target_path, disk)
                        vol = conn.storageVolLookupByPath(path)
                        vol.delete(0)
            if os.path.exists(output):
                shutil.rmtree(output)
    
    def create_tmp_folder(self, output):
        if not os.path.exists(output):
            os.makedirs(output, exist_ok=True)

    def subprocess_restore(self, command, error_messages, vm_name):
        result = subprocess.run(command, shell=True,
                                capture_output=True, text=True)
        messages = f"\n{result.stderr}"
        log.debug(messages)
        for message in messages.split('\n'):
            if "ERROR" in message:
                message = re.sub(
                    r'\x1b\[\d+m|\[\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\]',
                    '', message)
                error_messages.append(
                    (vm_name, message.split(":", 1)[1].strip()))

    def generate_guest(self, output):
        files = os.listdir(output)
        configname = list(filter(lambda file: file.endswith(".xml"), files))[0]
        conn = self.conn.get_backend()
        with open(os.path.join(output, configname), "r") as f:
            src_xml = f.read()
        _new_guest = Guest(conn, parsexml=src_xml)
        _new_guest.uuid = Guest.generate_uuid(conn)
        return _new_guest

    def existed_disks(self, disks, pool_name):
        conn = self.conn.get_backend()
        pool = conn.storagePoolLookupByName(pool_name)

        for disk in disks[:]:
            try:
                pool.storageVolLookupByName(disk)
            except:
                disks.remove(disk)
        return disks

    def save_disk(self, output, file_name, conn, pool_name, guest):
        pool = conn.storagePoolLookupByName(pool_name)
        spool = StoragePool(conn, pool.XMLDesc(0))
        # new_vol_name = _generate_restore_path(vm_name, new_name, file_name,
        #                                       lambda _: False)
        try: 
            vol = pool.storageVolLookupByName(file_name)
            raise Exception(_("Storage with name '%s' already exist")
                            % file_name)
        except:
            pass

        # Update target_path of disks: pool_path/new_volume_name
        for disk in guest.devices.disk:
            if disk.device == "cdrom":
                continue
            disk_name = os.path.basename(disk.source.file)
            if disk_name == file_name and disk.source.file:
                disk.source.file = os.path.join(spool.target_path, file_name)

        # Disk creation
        st = conn.newStream()
        cap = .000001

        vol = StorageVolume(conn)
        vol.pool = pool
        vol.name = file_name
        vol.capacity = (cap * 1024 * 1024 * 1024)
        vol.allocation = vol.capacity
        vol.install(meter=None)
        vol = pool.storageVolLookupByName(file_name)

        _upload_disk(vol, st, os.path.join(output, file_name))


    ################
    # UI listeners #
    ################

    def _close_cb(self, ignore1=None, ignore2=None):
        self.close()
        return 1

    def _finish_clicked_cb(self, src):
        ignore = src
        self._finish()

    def _browse_backup_local_clicked_cb(self, src):
        ingore = src
        self._browse_backup_folder()

    def _backup_folder_changed_cb(self, src):
        backuppath = src.get_text()
        if os.path.exists(backuppath):
            self._load_backup_folders(backuppath)
            self.config.set_default_directory("backup", backuppath)

    def _browse_restore_local_clicked_cb(self, src):
        ingore = src
        self._browse_restore_folder()

    def _action_type_changed_cb(self, src):
        action_type = uiutil.get_list_selection(src)
        self._action_type_changed(action_type)
