import datetime
import os
import shutil
import subprocess

from virtManager.asyncjob import vmmAsyncJob
from virtManager.lib import uiutil
from virtinst import generatename, log
from virtinst.devices.interface import DeviceInterface
from virtinst.guest import Guest
from virtinst.importer import _upload_disk
from virtinst.cloner import _get_shareable_msg
from virtinst.storage import StoragePool, StorageVolume
from .baseclass import vmmGObjectUI

def _generate_restore_path(origname, newname, origpath, cb_exists):
    """
    Generate desired replace path for auxiliary files, derived from the
    original path, original VM name, and proposed new VM name
    """
    if origpath is None:
        return None

    path = origpath
    suffix = ""

    # Try to split the suffix off the existing disk name. Ex.
    # foobar.img -> foobar-clone.img
    #
    # If the suffix is greater than 7 characters, assume it isn't
    # a file extension and is part of the disk name, at which point
    # just stick '-replace' on the end.
    if "." in origpath and len(origpath.rsplit(".", 1)[1]) <= 7:
        path, suffix = origpath.rsplit(".", 1)
        suffix = "." + suffix

    dirname = os.path.dirname(path)
    basename = os.path.basename(path)

    restorebase = basename.replace(origname, newname) 
    if restorebase == basename and origname not in restorebase:
        restorebase = basename + "-" + newname
    restorebase = os.path.join(dirname, restorebase)
    return generatename.generate_name(restorebase, cb_exists, suffix=suffix)


class vmmRestoreBackupDialog(vmmGObjectUI):
    _instance = None

    @classmethod
    def reset_instance(cls):
        cls._instance = None

    @classmethod
    def show_instance(cls, parentobj, vm):
        try:
            if cls._instance:
                cls._instance.close()
            cls.reset_instance()
            cls._instance = vmmRestoreBackupDialog(vm)
            cls._instance.show(parentobj.topwin, vm.conn)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(
                _("Error launching restore backup dialog: %s") % str(e))

    def __init__(self, vm):
        vmmGObjectUI.__init__(self, "restorebackup.ui", "vmm-restore-backup")
        self.vm = vm
        self._backup_folder = self.vm.get_xmlobj(inactive=True).get_backup_folder()
        self.output = None
        self.overwrite = False

        self.builder.connect_signals({
            "on_vmm_restore_delete_event": self._delete_event,
            "on_restore_cancel_clicked": self._cancel_clicked,
            "on_restore_finish_clicked": self._finish_clicked,
            "on_browse_local_clicked": self._browse_local_clicked,
        })
        self._init_backup_xml()
        self._populate_restore_combo(self.widget("restore-checkpoints"))
        self._populate_pool_combo(self.widget("pool-model"))
        self.widget("restore-name").set_text(self._new_guest.name)

    def _cleanup(self):
        self.vm = None
        self._backup_folder = None
        self.output = None

    def _init_backup_xml(self):
        files = os.listdir(self._backup_folder)
        config = sorted(list(filter(lambda file: file.endswith(".xml"), files)))
        if config:
            config_path = os.path.join(self._backup_folder, config[-1])
            conn = self.vm.conn.get_backend()
            with open(config_path, "r") as f:
                src_xml = f.read()
                self._new_guest = Guest(conn, parsexml=src_xml)


    ##############
    # Public API #
    ##############

    def show(self, parent, conn):
        log.debug("Showing restore wizard")
        self.conn = conn
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing restore wizard")
        self.topwin.hide()
        return 1

    def _vm_removed_cb(self, _conn, vm):
        if self.vm == vm:
            self.close()


    ###############
    # UI building #
    ###############

    def _populate_restore_combo(self, combo):
        if self._backup_folder:
            checkpoints = os.path.join(self._backup_folder, "checkpoints")
            items = []
            counter = 1
            for checkpoint in sorted(os.listdir(checkpoints), key=lambda x: (len(x), x)):
                path = os.path.join(checkpoints, checkpoint)
                m_time = os.path.getmtime(path)
                dt_m = datetime.datetime.fromtimestamp(m_time)
                view = f"{counter}: " + str(dt_m).rsplit(".", 1)[0]
                items.append([checkpoint, view])
                counter += 1

            items.sort(key=lambda x: int(x[1].split(":")[0]))
            uiutil.build_simple_combo(combo, items)

    def _populate_pool_combo(self, combo):
        backup_pool_path = None
        for disk in self._new_guest.devices.disk:
            backup_pool_path = os.path.dirname(disk.source.file)
            break

        pools = []
        default_pool = "default"
        for pool in self.vm.conn.list_pools():
            poolxml = pool.get_xmlobj()
            pools.append([None, poolxml.name])
            if poolxml.target_path == backup_pool_path:
                default_pool = poolxml.name
        uiutil.build_simple_combo(combo, pools, default_value=default_pool)


    #############
    # Listeners #
    #############

    def _delete_event(self, ignore1, ignore2):
        self.close()
        return 1

    def _cancel_clicked(self, src):
        ignore = src
        self.close()

    def _finish_clicked(self, src):
        ignore = src
        self._finish()

    def _browse_local_clicked(self, src):
        ingore = src
        self._browse_file()

    ####################
    # restore handling #
    ####################

    def _build_command(self):
        command = (f"virtnbdrestore -n -U {self.conn.get_uri()} -o {self.output} "
                   f"-i {self._backup_folder} --raw")

        checkpoint = uiutil.get_list_selection(self.widget("restore-checkpoints"))
        if checkpoint:
            checkpoint = checkpoint.rsplit(".", 1)[0]
            command += f" --until {checkpoint}"
        return command

    def _async_restore(self, asyncjob, new_name):
        self.output = os.path.join("/tmp/restore", self.vm.get_name())
        command = self._build_command()
        vm = None
        disks = []
        try:
            # Create temporary folder
            if not os.path.exists(self.output):
                os.makedirs(self.output, exist_ok=True)

            conn = self.conn.get_backend()
            pool_name = self.widget("pool-model-text").get_text()
            pool = conn.storagePoolLookupByName(pool_name)
            spool = StoragePool(conn, pool.XMLDesc(0))
            if not os.path.exists(spool.target_path):
                raise Exception(_("Pool path is not exists."))

            # Run virtnbdrestore
            result = subprocess.run(command, shell=True, capture_output=True, text=True)
            message = f"\n{result.stderr}"
            log.debug(message)
            encryption_error = "Encryption key was not found"
            if encryption_error in message:
                raise Exception(encryption_error)

            # Overwrite previous VM
            if self.overwrite:
                self.vm.delete()
                for disk in self.vm.xmlobj.devices.disk:
                    if not _get_shareable_msg(disk):
                        vol = conn.storageVolLookupByPath(disk.source.file)
                        vol.delete(0)

            # Create VM and upload disks
            files = os.listdir(self.output)
            config = list(filter(lambda file: file.endswith(".xml"), files))[0]
            self.save_xml(conn, new_name, config)
            for file in os.listdir(self.output):
                if not file.endswith('.xml'):
                    vol_name = self.save_disk(conn, pool, file, new_name)
                    disks.append(vol_name)
            vm = conn.defineXML(self._new_guest.get_xml())

            for pool in self.conn.list_pools():
                pool.refresh()

        except Exception as e:
            log.debug("Restoring failed: %s", str(e))
            if vm:
                vm.undefine()
            if disks:
                for disk in disks:
                    pool = StoragePool(conn, pool.XMLDesc(0))
                    path = os.path.join(pool.target_path, disk)
                    vol = conn.storageVolLookupByPath(path)
                    vol.delete(0)
            encryption_error = "Encryption key was not found"
            if encryption_error in str(e):
                raise Exception(_(encryption_error))
            raise Exception(e)
        finally:
            if os.path.exists(self.output):
                shutil.rmtree(self.output)

        log.debug("Restore finished.")

    def _finish(self):
        new_name = self.widget("restore-name").get_text()
        conn = self.conn.get_backend()
        dom = None
        try:
            dom = conn.lookupByName(new_name)
        except:
            pass
        if dom and new_name != self.vm.get_name():
            res = self.err.show_err(
                _("Virtual machine with the same name already exists"))
            return
        if new_name == self.vm.get_name():
            res = self.err.yes_no(
                _("Are you sure you want to restore virtual machine?"),
                _("Virtual machine with the same name already exists, "
                  "it will be overwritten. To create new backups you need to "
                  "delete old ones."))
            if not res:
                return
            self.overwrite = True
        else:
            res = self.err.yes_no(
                _("Are you sure you want to restore virtual machine?"),
                _("New virtual machine will be created"))
            if not res:
                return

        progWin = vmmAsyncJob(self._async_restore, [new_name], self._finish_cb,
                              [], _("Restoring virtual machine"),
                              _("This may take a while"), self.topwin)
        progWin.run()
        self.close()

    def _finish_cb(self, error, details):
        if error:
            error = (_("Error during virtual machine restore '%(vm)s': "
                        "%(error)s") % {
                        "vm": self.vm.get_name(), "error": error,
                        })
            self.err.show_err(error, details=details)

    def _init_new_guest(self, new_name, src_xml, conn):
        """
        Perform the series of unconditional new VM changes we always make
        """
        src_xml = src_xml.replace(self.vm.get_name(), new_name)
        self._new_guest = Guest(conn, parsexml=src_xml)
        self._new_guest.uuid = None
        self._new_guest.uuid = Guest.generate_uuid(conn)
        if not self.overwrite:
            self._new_guest.set_backup_schedule(None)
            self._new_guest.set_backup_repeat(None)
            self._new_guest.set_backup_folder(None)
            self._new_guest.set_backup_type(None)
            self._new_guest.set_backup_threshold(None)
            self._new_guest.set_backup_excluded(None)
            self._new_guest.set_backup_compress(None)
            self._new_guest.set_backup_included(None)
            self._new_guest.set_backup_encryption_key(None)
            self._new_guest.set_backup_hash(None)

        for iface in self._new_guest.devices.interface:
            iface.target_dev = None
            iface.macaddr = DeviceInterface.generate_mac(conn)

    def save_xml(self, conn, new_name, file_name):
        with open(os.path.join(self.output, file_name), "r") as f:
            src_xml = f.read()
            self._init_new_guest(new_name, src_xml, conn)

    def save_disk(self, conn, pool, file_name, new_name):
        vm_name = self.vm.get_name()
        spool = StoragePool(conn, pool.XMLDesc(0))
        new_vol_name = _generate_restore_path(vm_name, new_name, file_name,
                                              lambda _: False)
        try: 
            vol = pool.storageVolLookupByName(new_vol_name)
            raise Exception(_("Storage with name '%s' already exist")
                            % new_vol_name)
        except:
            pass

        # Update target_path of disks: pool_path/new_volume_name
        for disk in self._new_guest.devices.disk:
            if disk.device == "cdrom":
                continue
            disk_name = os.path.basename(disk.source.file)
            if disk_name == file_name and disk.source.file:
                disk.source.file = os.path.join(spool.target_path,
                                                new_vol_name)

        # Disk creation
        st = conn.newStream()
        cap = .000001

        vol = StorageVolume(conn)
        vol.pool = pool
        vol.name = new_vol_name
        vol.capacity = (cap * 1024 * 1024 * 1024)
        vol.allocation = vol.capacity
        vol.install(meter=None)
        vol = pool.storageVolLookupByName(new_vol_name)

        _upload_disk(vol, st, os.path.join(self.output, file_name))
        return new_vol_name
