# Copyright (C) 2009, 2013, 2014 Red Hat, Inc.
# Copyright (C) 2009 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
import logging
import re
from gi.repository import Gtk

from virtinst import log

from .asyncjob import vmmAsyncJob


####################################################################
# Build toolbar shutdown button menu (manager and details toolbar) #
####################################################################

class _VMMenu(Gtk.Menu):
    def __init__(self, src, current_vm_cb, show_open=True):
        Gtk.Menu.__init__(self)
        self._parent = src
        self._current_vm_cb = current_vm_cb
        self._show_open = show_open

        self._init_state()

    def _add_action(self, label, widgetname, cb, tooltip=None):
        item = Gtk.MenuItem.new_with_mnemonic(label)

        item.vmm_widget_name = widgetname
        if cb:
            def _cb(_menuitem):
                _vm = self._current_vm_cb()
                if _vm:
                    return cb(self._parent, _vm)
            item.connect("activate", _cb)
        if tooltip:
            item.set_tooltip_text(tooltip)
        self.add(item)
        return item

    def _init_state(self):
        raise NotImplementedError()
    def update_widget_states(self, vm):
        raise NotImplementedError()


class VMShutdownMenu(_VMMenu):
    """
    Shutdown submenu for reboot, forceoff, reset, etc.
    """
    def _init_state(self):
        self._add_action(_("_Reboot"), "reboot", VMActionUI.reboot)
        self._add_action(_("_Shut Down"), "shutdown", VMActionUI.shutdown)
        self._add_action(_("F_orce Reset"), "reset", VMActionUI.reset)
        self._add_action(_("_Force Off"), "destroy", VMActionUI.destroy)
        self.add(Gtk.SeparatorMenuItem())
        self._add_action(_("Sa_ve"), "save", VMActionUI.save)

        self.get_accessible().set_name("vmm-shutdown-menu")
        self.show_all()

    def update_widget_states(self, vm):
        statemap = {
            "reboot": bool(vm and vm.is_stoppable()),
            "shutdown": bool(vm and vm.is_stoppable()),
            "reset": bool(vm and vm.is_stoppable()),
            "destroy": bool(vm and vm.is_destroyable()),
            "save": bool(vm and vm.is_destroyable()),
        }

        for child in self.get_children():
            name = getattr(child, "vmm_widget_name", None)
            if name in statemap:
                child.set_sensitive(statemap[name])


class VMBackupMenu(_VMMenu):
    """
    Backup submenu: backup, restore
    """
    def _init_state(self):
        self._add_action(_("Backup..."), "backup", VMActionUI.backup,
                         tooltip=self.backup_tooltip())
        self._add_action(_("Restore..."), "restore_backup",
                         VMActionUI.restore_backup,
                         tooltip=self.restore_tooltip())
        self._add_action(_("Integrity control"), "backup_integrity",
                         VMActionUI.backup_integrity)

    def update_widget_states(self, vm):
        statemap = {
            "backup": bool(vm and vm.is_backupable()),
            "restore_backup": bool(vm and vm.is_restorable()),
            "backup_integrity": bool(
                vm and vm.get_xmlobj(inactive=True).get_backup_folder())
        }

        for child in self.get_children():
            name = getattr(child, "vmm_widget_name", None)
            if name in statemap:
                child.set_sensitive(statemap[name])

    def backup_tooltip(self):
        def check_package(package_name):
            import subprocess
            try:
                subprocess.check_output(['dpkg-query', '-W', '-f=${Status}', package_name])
                return True
            except subprocess.CalledProcessError:
                return False

        tooltip = ""
        if not check_package("virtnbdbackup"):
            tooltip += "Virtnbdbackup isn't installed"
        if not tooltip:
            return None
        return _(tooltip)

    def restore_tooltip(self):
        return _("The virtual machine must be shutted down.")


class VMActionMenu(_VMMenu):
    """
    VM submenu for run, pause, shutdown, clone, etc
    """
    def _init_state(self):
        self._add_action(_("_Run"), "run", VMActionUI.run)
        self._add_action(_("_Pause"), "suspend", VMActionUI.suspend)
        self._add_action(_("R_esume"), "resume", VMActionUI.resume)
        s = self._add_action(_("_Shut Down"), "shutdown", None)
        s.set_submenu(VMShutdownMenu(self._parent, self._current_vm_cb))

        self.add(Gtk.SeparatorMenuItem())
        b = self._add_action(_("Backup"), "backup", None)
        b.set_submenu(VMBackupMenu(self._parent, self._current_vm_cb))
        self._add_action(_("Clone..."), "clone", VMActionUI.clone)
        self._add_action(_("Migrate..."), "migrate", VMActionUI.migrate)
        self._add_action(_("Export..."), "export", VMActionUI.export)
        self._add_action(_("_Delete"), "delete", VMActionUI.delete)

        if self._show_open:
            self.add(Gtk.SeparatorMenuItem())
            self._add_action(_("_Open"), "show", VMActionUI.show)

        self.add(Gtk.SeparatorMenuItem())
        self._add_action(_("Add To Group"), "add_to_group",
                         VMActionUI.add_to_group)
        self._add_action(_("Remove From Group"), "remove_from_group",
                         VMActionUI.remove_from_group)

        self.get_accessible().set_name("vm-action-menu")
        self.show_all()

    def update_widget_states(self, vm):
        statemap = {
            "run": bool(vm and vm.is_runable()),
            "shutdown": bool(vm and vm.is_stoppable()),
            "suspend": bool(vm and vm.is_stoppable()),
            "resume": bool(vm and vm.is_paused()),
            "migrate": bool(vm and vm.is_stoppable()),
            "export": bool(vm and vm.is_shutoff()),
            "clone": bool(vm and vm.is_cloneable()),
            "add_to_group": bool(vm),
            "remove_from_group": bool(vm and vm.is_ungroupable()),
        }
        vismap = {
            "suspend": bool(vm and not vm.is_paused()),
            "resume": bool(vm and vm.is_paused()),
        }

        for child in self.get_children():
            name = getattr(child, "vmm_widget_name", None)
            if child.get_submenu():
                child.get_submenu().update_widget_states(vm)
            if name in statemap:
                child.set_sensitive(statemap[name])
            if name in vismap:
                child.set_visible(vismap[name])

    def change_run_text(self, text):
        for child in self.get_children():
            if getattr(child, "vmm_widget_name", None) == "run":
                child.get_child().set_label(text)


class VMActionUI(object):
    """
    Singleton object for handling VM actions, asking for confirmation,
    showing errors/progress dialogs, etc.
    """

    @staticmethod
    def save_cancel(asyncjob, vm):
        log.debug("Cancelling save job")
        if not vm:
            return  # pragma: no cover

        try:
            vm.abort_job()
        except Exception as e:
            log.exception("Error cancelling save job")
            asyncjob.show_warning(_("Error cancelling save job: %s") % str(e))
            return

        asyncjob.job_canceled = True  # pragma: no cover

    @staticmethod
    def save(src, vm):
        if not src.err.chkbox_helper(src.config.get_confirm_poweroff,
                src.config.set_confirm_poweroff,
                text1=_("Are you sure you want to save '%s'?") % vm.get_name()):
            return

        _cancel_cb = None
        if vm.supports_domain_job_info():
            _cancel_cb = (VMActionUI.save_cancel, vm)

        def cb(asyncjob):
            vm.save(meter=asyncjob.get_meter())
        def finish_cb(error, details):
            if error is not None:  # pragma: no cover
                error = _("Error saving domain: %s") % error
                src.err.show_err(error, details=details)

        progWin = vmmAsyncJob(cb, [],
                    finish_cb, [],
                    _("Saving Virtual Machine"),
                    _("Saving virtual machine memory to disk "),
                    src.topwin, cancel_cb=_cancel_cb)
        progWin.run()

    @staticmethod
    def destroy(src, vm):
        if not src.err.chkbox_helper(
            src.config.get_confirm_forcepoweroff,
            src.config.set_confirm_forcepoweroff,
            text1=_("Are you sure you want to force poweroff '%s'?") %
                    vm.get_name(),
            text2=_("This will immediately poweroff the VM without "
                    "shutting down the OS and may cause data loss.")):
            return

        log.debug("Destroying vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.destroy, [], src,
                                        _("Error shutting down domain"))

    @staticmethod
    def suspend(src, vm):
        if not src.err.chkbox_helper(src.config.get_confirm_pause,
            src.config.set_confirm_pause,
            text1=_("Are you sure you want to pause '%s'?") %
                    vm.get_name()):
            return

        log.debug("Pausing vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.suspend, [], src,
                                        _("Error pausing domain"))

    @staticmethod
    def resume(src, vm):
        log.debug("Unpausing vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.resume, [], src,
                                        _("Error unpausing domain"))

    @staticmethod
    def run(src, vm):
        def ensure_network_active():
            interfaces = vm.get_interface_devices_norefresh()
            if len(interfaces):
                iface = interfaces[0]
            else:
                src.err.show_err(_("No interfaces in virtual machine: %s")
                                 % vm.get_name())
                return
            netobj = vm.conn.get_net_by_name(iface.network)
            if not netobj or netobj.is_active():
                return

            res = src.err.yes_no(_("Virtual Network is not active."),
                _("Virtual Network '%s' is not active. "
                "Would you like to start the network now?") % iface.network)
            if not res:
                return  # pragma: no cover
            netobj.start()

        # rbt: restore ro image
        def restore_ro_image(vm, title):

            def cb(asyncjob):
                vm.startup()

            def finish_cb(error, details):
                if error is not None:
                    error = _("Error starting domain: %s") % error
                    src.err.show_err(error, details=details)

            logging.debug('Restoring ro image')
            job = vmmAsyncJob(cb, [],
                              finish_cb, [],
                              title,
                              _("Please wait, while libvirt restore read only image"),
                              src.topwin)
            job.run()

        log.debug("Starting vm '%s'", vm.get_name())

        if vm.has_managed_save():
            def errorcb(error, details):
                # This is run from the main thread
                res = src.err.show_err(
                    _("Error restoring domain: %s") % error,
                    details=details,
                    text2=_(
                        "The domain could not be restored. Would you like\n"
                        "to remove the saved state and perform a regular\n"
                        "start up?"),
                    dialog_type=Gtk.MessageType.WARNING,
                    buttons=Gtk.ButtonsType.YES_NO,
                    modal=True)

                if not res:
                    return

                try:
                    vm.remove_saved_image()
                    VMActionUI.run(src, vm)
                except Exception as e:  # pragma: no cover
                    src.err.show_err(_("Error removing domain state: %s")
                                     % str(e))

            # VM will be restored, which can take some time, so show progress
            title = _("Restoring Virtual Machine")
            text = _("Restoring virtual machine memory from disk")
            # rbt:
            if vm.get_readonly():
                restore_ro_image(vm, title)
            else:
                logging.debug('Restoring not ro image')
                vmmAsyncJob.simple_async(vm.startup, [], src,
                                     title, text, "", errorcb=errorcb)

        else:
            # rbt: restore regular ro image
            if vm.get_readonly():
                restore_ro_image(vm, _("Restoring Virtual Machine"))
            else:
                # Regular startup
                errorintro  = _("Error starting domain")
                ensure_network_active()
                vmmAsyncJob.simple_async_noshow(vm.startup, [], src, errorintro)

    @staticmethod
    def shutdown(src, vm):
        if not src.err.chkbox_helper(src.config.get_confirm_poweroff,
            src.config.set_confirm_poweroff,
            text1=_("Are you sure you want to poweroff '%s'?") %
                    vm.get_name()):
            return

        log.debug("Shutting down vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.shutdown, [], src,
                                        _("Error shutting down domain"))

    @staticmethod
    def reboot(src, vm):
        if not src.err.chkbox_helper(src.config.get_confirm_poweroff,
            src.config.set_confirm_poweroff,
            text1=_("Are you sure you want to reboot '%s'?") %
                    vm.get_name()):
            return

        log.debug("Rebooting vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.reboot, [], src,
            _("Error rebooting domain"))

    @staticmethod
    def reset(src, vm):
        if not src.err.chkbox_helper(
            src.config.get_confirm_forcepoweroff,
            src.config.set_confirm_forcepoweroff,
            text1=_("Are you sure you want to force reset '%s'?") %
                    vm.get_name(),
            text2=_("This will immediately reset the VM without "
                    "shutting down the OS and may cause data loss.")):
            return

        log.debug("Resetting vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.reset, [], src,
                                        _("Error resetting domain"))

    @staticmethod
    def delete(src, vm):
        from .delete import vmmDeleteDialog
        vmmDeleteDialog.show_instance(src, vm)

    @staticmethod
    def migrate(src, vm):
        from .migrate import vmmMigrateDialog
        vmmMigrateDialog.show_instance(src, vm)

    @staticmethod
    def clone(src, vm):
        from .clone import vmmCloneVM
        vmmCloneVM.show_instance(src, vm)

    @staticmethod
    def export(src, vm):
        from .export import vmmExportVM
        vmmExportVM.show_instance(src, vm)

    @staticmethod
    def backup(src, vm):
        from .backup import vmmBackupDialog
        vmmBackupDialog.show_instance(src, vm)

    @staticmethod
    def restore_backup(src, vm):
        from .restorebackup import vmmRestoreBackupDialog
        vmmRestoreBackupDialog.show_instance(src, vm)

    @staticmethod
    def backup_integrity(src, vm):
        import subprocess
        command = (f"virtnbdrestore -i {vm.get_xmlobj().get_backup_folder()} "
                   "-o verify")
        error_messages = []

        def _async_verify(command, error_messages):
            result = subprocess.run(command, shell=True, capture_output=True,
                                    text=True)
            messages = f"\n{result.stderr}"
            log.debug(messages)

            messages = messages.split("\n")
            for message in messages:
                if "ERROR" in message:
                    message = re.sub(
                        r'\x1b\[\d+m|\[\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\]',
                        '', message)
                    log.debug(message)
                    error_messages.append(message.split(":", 1)[1].strip())

        def finish_cb():
            def error_view(error):
                parts = error.split(":")
                message = _(parts[0])
                return (_("%(message)s in '%(vm)s':%(comparement)s") %
                        {"message": message, "vm": vm.get_name(),
                         "comparement": parts[1]})
            nonlocal error_messages
            if error_messages:
                error_messages = [error_view(e) for e in error_messages]
                src.err.show_err(error_messages[0], "\n".join(error_messages))
            else:
                src.err.show_info(_("All checksums match"))

        title = _("Verification")
        text = _("Creation and verification of checksums")
        vmmAsyncJob.simple_async(_async_verify, [command, error_messages], src,
                                 title, text, "", finish_cb=finish_cb)

    @staticmethod
    def show(src, vm):
        from .vmwindow import vmmVMWindow
        vmmVMWindow.get_instance(src, vm).show()

    @staticmethod    
    def add_to_group(src, vm):
        from .group import vmmGroupDialog
        vmmGroupDialog.show_instance(src, vm)
    
    @staticmethod    
    def remove_from_group(src, vm):
        vm.delete_vm_group()
